import functools
from typing import Callable, Optional

import structlog

from psenv.core import config

from . import exceptions


class ErrorHandler:
    def __init__(self, name: Optional[str] = None) -> None:
        self.name = name
        self.raise_errors = config.PSENV_RAISE_ERRORS
        self.logger = structlog.get_logger(name or self.__class__.__name__)

    def handle(self, exception: Exception) -> int:
        if self.raise_errors:
            raise exception
        self.logger.error(exception)
        return 1


def handle_cli_errors(func: Callable) -> Callable:
    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> int:
        error_handler = ErrorHandler()

        try:
            return func(*args, **kwargs)

        except exceptions.PsenvParameterStoreError as e:
            return error_handler.handle(e)

        except exceptions.PsenvConfigNotFoundError as e:
            return error_handler.handle(e)

        except exceptions.PsenvConfigError as e:
            return error_handler.handle(e)

    return wrapper



import os
from io import StringIO
from pathlib import Path
from typing import Any, Dict, Optional, TextIO

import yaml
from dotenv import dotenv_values

from psenv.core.config import PSENV_PRIVATE_MARKER
from psenv.core.paths import PSENV_TEMPLATE_FILE_PATH
from psenv.error_handling.exceptions import PsenvConfigError, PsenvConfigNotFoundError, PsenvInternalError


def read_config(path: Path) -> Dict[str, Any]:
    path = path.absolute()
    try:
        content = os.path.expandvars(path.read_text())
        return yaml.safe_load(content)
    except FileNotFoundError:
        raise PsenvConfigNotFoundError(
            f"psenv.yml not found in {path.as_posix()}. Please check the path or create a new configuration file."
        )
    except yaml.YAMLError as e:
        raise PsenvConfigError("Error loading psenv.yml") from e


def read_config_template(path: Optional[Path] = None) -> str:
    path = path or PSENV_TEMPLATE_FILE_PATH
    try:
        return path.absolute().read_text()
    except FileNotFoundError:
        raise PsenvInternalError(f"psenv template file not found at {path}. Please file a bug report.")


def write_config(path: Path, content: str) -> None:
    path = path.absolute()
    path.parent.mkdir(parents=True, exist_ok=True)
    path.touch(exist_ok=True)
    path.write_text(content)


def get_environment_variables(prefix: str) -> Dict[str, str]:
    retrieved = {}
    prefix = prefix.upper()
    for key, value in os.environ.items():
        if key.startswith(prefix):
            retrieved[key] = value
    return retrieved


class EnvFile:
    def __init__(self, path: Path) -> None:
        self.path = path.expanduser().absolute()
        self._main_params = {}
        self._private_params = {}

        if not self.path.exists():
            self.path.parent.mkdir(parents=True, exist_ok=True)
            self.path.touch(exist_ok=True)

    @property
    def local_params(self) -> Dict[str, str]:
        return self._main_params

    @staticmethod
    def _split_sections(content: str) -> tuple[str, str]:
        mp, _, pp = content.partition(PSENV_PRIVATE_MARKER)
        return mp, pp

    @staticmethod
    def _parse_env_section(section: str) -> Dict[str, str]:
        with StringIO() as stream:
            stream.write(section)
            stream.seek(0)
            return dotenv_values(stream=stream)

    @staticmethod
    def _get_prefix(key: str) -> str:
        return key.split("_")[0]

    @staticmethod
    def _write_section_params(params: Dict[str, str], env: TextIO) -> None:
        last_prefix = None
        for key in sorted(params.keys()):
            prefix = EnvFile._get_prefix(key)
            if prefix != last_prefix:
                env.write("\n")
                env.write(f"# ----------------------------{prefix}------------------------------------ #\n")
                last_prefix = prefix
            env.write(f"{key}='{params[key]}'\n")

    def load(self) -> None:
        content = self.path.read_text()
        mp, pp = self._split_sections(content)
        self._main_params = self._parse_env_section(mp)
        self._private_params = self._parse_env_section(pp)

    def get_params(self, section: str) -> Dict[str, str]:
        return getattr(self, f"_{section}_params", {})

    def update_params(self, params: Dict[str, str], section: str) -> None:
        p = self.get_params(section)
        p.update(**params)
        setattr(self, f"_{section}_params", p)

    def write_params(self) -> None:
        with self.path.open(mode="w+") as env:
            self._write_section_params(self._main_params, env)
            env.write(f"\n\n{PSENV_PRIVATE_MARKER}\n\n")
            self._write_section_params(self._private_params, env)



from argparse import Namespace
from pathlib import Path
from typing import Any

import structlog

from psenv.core.fileio import EnvFile, get_environment_variables
from psenv.core.models import load_config

logger = structlog.get_logger(__name__)


def configure_parser(subparser: Any) -> None:
    inject_parser = subparser.add_parser(
        name="inject",
        description="Inject environment variables from your session to an .env file",
        help="Inject environment variables to an .env file",
    )
    inject_parser.set_defaults(func=inject_parameters)

    inject_parser.add_argument("-e", "--env", type=str, required=False, help="The environment to inject parameters to.")
    inject_parser.add_argument(
        "-p", "--prefix", type=str, required=False, default="", help="The prefix to filter environment variables by."
    )


def inject_parameters(cliargs: Namespace) -> None:
    logger.info("Injecting parameters into env file:", config=cliargs.config)

    config = load_config(cliargs.config)

    if cliargs.env:
        config_env = config.get_config_environment(cliargs.env)
        env_file = EnvFile(Path(config_env.environment.envfile))
    else:
        env_file = EnvFile(Path(config.envfile))

    env_file.load()
    env_vars = get_environment_variables(cliargs.prefix)
    env_file.update_params(env_vars, section="private")
    env_file.write_params()



from argparse import Namespace
from typing import Any

import structlog

from psenv.core import diff
from psenv.core.context import Context
from psenv.core.synchronizer import Synchronizer

logger = structlog.get_logger(__name__)


def configure_parser(subparser: Any) -> None:
    put_parser = subparser.add_parser(
        name="put", description="Manage psenv configurations.", help="Put parameters into the parameter store."
    )
    put_parser.set_defaults(func=put_parameters)

    put_parser.add_argument("-e", "--env", type=str, required=True, help="The environment to put parameters for.", metavar="")

    group = put_parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-a", "--add", action="store_true", help="Add new parameters only.")
    group.add_argument("-u", "--update", action="store_true", help="Add new and update existing parameters.")
    group.add_argument(
        "-s", "--sync", action="store_true", help="Add new, update existing, and remove parameters not in the local environment file."
    )

    put_parser.add_argument("--dry-run", action="store_true", help="Show what would be done, but do not make any changes.")


def put_parameters(cliargs: Namespace) -> None:
    logger.info("Putting parameters into the parameter store for environment:", config=cliargs.config)

    ctx = Context.from_cliargs(cliargs)

    # fetch the remote parameters from the parameter store
    remote_params = ctx.ps_client.get_parameters()

    # generate a parameter diff to use for the put operation
    param_diff = diff.diff_parameters(ctx.env_file.local_params, remote_params)

    # get the synchronizer
    synchronizer = Synchronizer(ctx=ctx, param_diff=param_diff, mode=Synchronizer.get_mode_from_cliargs(cliargs), dry_run=cliargs.dry_run)
    # execute the synchronization
    synchronizer.sync()



from typing import Dict


class ParameterDiff:
    def __init__(self, to_add: Dict[str, str], to_update: Dict[str, str], to_remove: Dict[str, str]) -> None:
        self.to_add = to_add
        self.to_update = to_update
        self.to_remove = to_remove


def diff_parameters(local_parameters: Dict[str, str], remote_parameters: Dict[str, str]) -> ParameterDiff:
    to_add = {}
    to_update = {}
    to_remove = {}

    for key, value in local_parameters.items():
        if key in remote_parameters.keys():
            if value != remote_parameters[key]:
                to_update[key] = value
        else:
            to_add[key] = value

    for key, value in remote_parameters.items():
        if key not in local_parameters:
            to_remove[key] = value

    return ParameterDiff(to_add=to_add, to_update=to_update, to_remove=to_remove)



from pathlib import Path

PSENV_FILE_NAME = "psenv.yml"
PSENV_CONFIG_FILE_PATH = Path.cwd() / PSENV_FILE_NAME
PSENV_TEMPLATE_FILE_PATH = Path(__file__).parent.parent / "templates" / "psenv.template.yml"



from typing import Optional

import boto3
from mypy_boto3_sts import STSClient
from mypy_boto3_sts.type_defs import GetCallerIdentityResponseTypeDef

from psenv.error_handling.exceptions import PsenvInternalError, PsenvInvalidAccountError


class StsClient:
    def __init__(self, sts_client: Optional[STSClient] = None) -> None:
        self._sts = sts_client or boto3.client("sts")
        self._caller_identity: Optional[GetCallerIdentityResponseTypeDef] = None

    @property
    def sts(self) -> STSClient:
        return self._sts

    @property
    def caller_identity(self) -> GetCallerIdentityResponseTypeDef:
        if not self._caller_identity:
            self._caller_identity = self.get_caller_identity()
        return self._caller_identity

    @property
    def account_id(self) -> str:
        return self.caller_identity.get("Account", "")

    def get_caller_identity(self) -> GetCallerIdentityResponseTypeDef:
        try:
            return self.sts.get_caller_identity()
        except Exception as e:
            raise PsenvInternalError(f"Failed to get caller identity: {e}") from e

    def raise_if_invalid_account(self, expected_account_id: str) -> None:
        if self.account_id != expected_account_id:
            raise PsenvInvalidAccountError(f"Invalid AWS account ID: {self.account_id}. Expected: {expected_account_id}.")



import structlog

from psenv.core.context import Context
from psenv.core.diff import ParameterDiff
from psenv.error_handling.exceptions import PsenvCLIError


class Synchronizer:
    def __init__(self, ctx: Context, param_diff: ParameterDiff, mode: str, dry_run: bool) -> None:
        self.ctx = ctx
        self.param_diff = param_diff
        self.mode = mode
        self.dry_run = dry_run
        self.logger = structlog.getLogger("psenv.synchronizer")

    def sync(self):
        if self._is_nothing_to_do():
            return

        if self.dry_run:
            self.execute_dry_run()
            return

        self.execute()

    def execute_dry_run(self) -> None:
        if self.mode == "add":
            self._log_dry_run_to_add()

        elif self.mode == "update":
            self._log_dry_run_to_add()
            self._log_dry_run_to_update()

        elif self.mode == "sync":
            self._log_dry_run_to_add()
            self._log_dry_run_to_update()
            self._log_dry_run_to_remove()
        else:
            raise PsenvCLIError(f"Invalid mode: {self.mode}. Must be one of 'add', 'update', or 'sync'.")

    def execute(self) -> None:
        if self.mode == "add":
            self._add_parameters()

        elif self.mode == "update":
            self._add_parameters()
            self._update_parameters()

        elif self.mode == "sync":
            self._add_parameters()
            self._update_parameters()
            self._delete_parameters()
        else:
            raise PsenvCLIError(f"Invalid mode: {self.mode}. Must be one of 'add', 'update', or 'sync'.")

    def _is_nothing_to_do(self) -> bool:
        if not self.param_diff.to_add and not self.param_diff.to_update and not self.param_diff.to_remove:
            self.logger.info("No parameters to add, update, or remove. Everything is up to date.")
            return True
        return False

    def _add_parameters(self) -> None:
        self.ctx.ps_client.put_parameters(self.param_diff.to_add)

    def _update_parameters(self) -> None:
        self.ctx.ps_client.put_parameters(self.param_diff.to_update, overwrite=True)

    def _delete_parameters(self) -> None:
        self.ctx.ps_client.delete_parameters(self.param_diff.to_remove)

    def _log_dry_run_to_add(self) -> None:
        for key in self.param_diff.to_add.keys():
            self.logger.info("[DRY RUN] Would add parameters: %s", key)

    def _log_dry_run_to_update(self) -> None:
        for key in self.param_diff.to_update.keys():
            self.logger.info("[DRY RUN] Would update parameters: %s", key)

    def _log_dry_run_to_remove(self) -> None:
        for key in self.param_diff.to_remove.keys():
            self.logger.info("[DRY RUN] Would remove parameters: %s", key)

    @staticmethod
    def get_mode_from_cliargs(cliargs) -> str:
        return Synchronizer.get_mode_from_flags(cliargs.add, cliargs.update, cliargs.sync)

    @staticmethod
    def get_mode_from_flags(add: bool, update: bool, sync: bool) -> str:
        flags = [add, update, sync]
        if sum(flags) != 1:
            raise ValueError("Exactly one of --add, --update, or --sync must be set.")
        if add:
            return "add"
        if update:
            return "update"
        if sync:
            return "sync"
        raise PsenvCLIError("Invalid mode flags.")



from argparse import Action, Namespace
from pathlib import Path
from typing import Any

import structlog

from psenv.core import fileio
from psenv.core.paths import PSENV_CONFIG_FILE_PATH
from psenv.error_handling.exceptions import PsenvConfigError

logger = structlog.get_logger(__name__)


class ValidatePathAction(Action):
    def __call__(self, parser, namespace, values: Path, option_string=None):
        if not values.suffix:
            raise PsenvConfigError(f"Path is not a file: {values}")
        if not values.suffix == ".yml":
            raise PsenvConfigError(f"Invalid file type: {values.suffix}. Expected a .yml file.")
        if values.exists():
            raise PsenvConfigError(f"Psenv config file already exists: {values}")
        setattr(namespace, self.dest, values)


def configure_parser(subparser: Any) -> None:
    config_parser = subparser.add_parser(name="config", description="Manage psenv configurations.")

    subparser = config_parser.add_subparsers(dest="new_config")
    new_parser = subparser.add_parser(name="new", description="Create a new psenv configuration file.")
    new_parser.set_defaults(func=new_config)
    new_parser.add_argument(
        "-p",
        "--path",
        help="Path to the new configuration file.",
        type=Path,
        required=False,
        default=PSENV_CONFIG_FILE_PATH,
        action=ValidatePathAction,
    )


def new_config(cliargs: Namespace) -> int:
    logger.info("Creating new psenv configuration file", path=cliargs.path)
    template = fileio.read_config_template()
    fileio.write_config(cliargs.path, template)
    return 0



from argparse import Namespace
from pathlib import Path

from dotenv import load_dotenv

from psenv import aws
from psenv.core import fileio, models


class Context:
    def __init__(self, env: str, config: Path) -> None:
        # Load the configuration file and get the environment configuration
        self.config = models.load_config(config)
        self.config_env = self.config.get_config_environment(env)

        # Load environment variables from the .env file specified in the config
        load_dotenv(self.config.envfile)

        # Check the sts account to ensure we are in the expected AWS account
        self.sts = aws.StsClient()
        self.sts.raise_if_invalid_account(self.config_env.environment.account)

        # Load the environment file and its parameters we are using for this context
        self.env_file = fileio.EnvFile(Path(self.config_env.environment.envfile))
        self.env_file.load()

        # Initialize the Parameter Store client with the parameter path and KMS key from the config
        self.ps_client = aws.ParameterStoreClient(parameter_path=self.config_env.parameter_path, kms_key=self.config_env.kms_key)

    @classmethod
    def from_cliargs(cls, cliargs: Namespace) -> "Context":
        """Create a Context instance from command line arguments."""
        return cls(env=cliargs.env, config=cliargs.config)



from typing import Any, Dict, Iterable, Optional, Tuple

import boto3
import structlog
from mypy_boto3_ssm import SSMClient

from psenv.error_handling.exceptions import PsenvParameterStoreError


class ParameterStoreClient:
    def __init__(self, parameter_path: str, ssm_client: Optional[SSMClient] = None, kms_key: Optional[str] = None) -> None:
        self._parameter_path = parameter_path
        self._ssm = ssm_client or boto3.client("ssm")
        self._kms_key = kms_key
        self.logger = structlog.getLogger("psenv.parameter_store")

    @property
    def ssm(self) -> SSMClient:
        return self._ssm

    @property
    def parameter_path(self) -> str:
        return self._parameter_path

    @property
    def kms_key(self) -> str:
        return self._kms_key or "alias/aws/ssm"

    @property
    def get_params_kwargs(self) -> Dict[str, Any]:
        return {
            "Path": self.parameter_path,
            "Recursive": True,
            "WithDecryption": True,
        }

    def put_params_kwargs(self, name: str, value: str, overwrite: bool) -> Dict[str, Any]:
        return {
            "Name": f"{self.parameter_path}/{name}",
            "Value": value,
            "Type": "SecureString",
            "Overwrite": overwrite,
            "KeyId": self.kms_key,
        }

    @staticmethod
    def parse_parameter_name(name: str) -> str:
        return name.split("/")[-1].upper()

    def _get_parameters(self) -> Iterable[Tuple[str, str]]:
        paginator = self.ssm.get_paginator("get_parameters_by_path")
        try:
            for page in paginator.paginate(**self.get_params_kwargs):
                for parameter in page.get("Parameters", []):
                    name = self.parse_parameter_name(parameter["Name"])
                    yield name, parameter["Value"]
        except Exception as e:
            raise PsenvParameterStoreError(f"Error fetching parameters {e}") from e

    def get_parameters(self) -> Dict[str, str]:
        return dict(self._get_parameters())

    def put_parameters(self, parameters: Dict[str, str], overwrite: bool = False) -> None:
        for name, value in parameters.items():
            try:
                self.logger.info("Putting parameter", name=f"{self.parameter_path}/{name}", overwrite=overwrite)
                self.ssm.put_parameter(**self.put_params_kwargs(name, value, overwrite))
            except Exception as e:
                raise PsenvParameterStoreError(f"Error putting Parameters: {e}") from e

    def delete_parameters(self, parameters: Dict[str, str]) -> None:
        for name in parameters:
            try:
                self.logger.info("Deleting parameter", name=f"{self.parameter_path}/{name}")
                self.ssm.delete_parameter(Name=f"{self.parameter_path}/{name}")
            except Exception as e:
                raise PsenvParameterStoreError(f"Error deleting parameter {name}: {e}") from e



from pathlib import Path
from typing import List, Optional

from pydantic import BaseModel, ValidationError, field_validator

from psenv.core.fileio import read_config
from psenv.core.paths import PSENV_CONFIG_FILE_PATH
from psenv.error_handling.exceptions import PsenvConfigError


class Environment(BaseModel):
    name: str
    account: str
    envfile: str
    kms_key: Optional[str] = None
    path: Optional[str] = None

    @field_validator("account")
    def validate_account(cls, value: str) -> str:
        if not value.isdigit() or len(value) != 12:
            raise PsenvConfigError("AWS account ID must be a 12-digit numeric string.")
        return value

    @field_validator("path")
    def validate_path(cls, value: Optional[str]) -> Optional[str]:
        if value and not value.startswith("/"):
            raise PsenvConfigError("Path must start with '/'.")
        return value

    @field_validator("kms_key")
    def validate_kms_key(cls, value: Optional[str]) -> Optional[str]:
        if value and not value.startswith("alias/"):
            raise PsenvConfigError("KMS key must start with 'alias/'.")
        return value


class PsenvConfig(BaseModel):
    envfile: str
    root_path: str
    root_kms_key: Optional[str] = None
    environments: List[Environment]

    @field_validator("root_path")
    def validate_path(cls, value: str) -> Optional[str]:
        if not value.startswith("/"):
            raise PsenvConfigError("Path must start with '/'.")
        return value

    @field_validator("root_kms_key")
    def validate_kms_key(cls, value: Optional[str]) -> Optional[str]:
        if value and not value.startswith("alias/"):
            raise PsenvConfigError("KMS key must start with 'alias/'.")
        return value

    def get_config_environment(self, name: str) -> "ConfigEnvironment":
        for env in self.environments:
            if env.name == name:
                return ConfigEnvironment(config=self, environment=env)
        raise PsenvConfigError(f"Environment '{name}' not found in configuration.")


class ConfigEnvironment(BaseModel):
    config: PsenvConfig
    environment: Environment

    @property
    def parameter_path(self) -> str:
        if self.environment.path:
            return f"{self.config.root_path}/{self.environment.name}{self.environment.path}"
        return self.config.root_path

    @property
    def kms_key(self) -> Optional[str]:
        return self.environment.kms_key or self.config.root_kms_key


def load_config(path: Optional[Path] = None) -> PsenvConfig:
    path = path or PSENV_CONFIG_FILE_PATH
    config = read_config(path)
    try:
        return PsenvConfig.model_validate(config)
    except (ValidationError, AssertionError, PsenvConfigError) as e:
        raise PsenvConfigError(f"Error loading config {e}")



from argparse import Namespace
from typing import Any

import structlog

from psenv.core.context import Context

logger = structlog.get_logger(__name__)


def configure_parser(subparser: Any) -> None:
    get_parser = subparser.add_parser(
        name="get", description="Manage psenv configurations.", help="Get parameters from the parameter store."
    )
    get_parser.set_defaults(func=get_parameters)

    get_parser.add_argument(
        "-e",
        "--env",
        type=str,
        required=True,
    )


def get_parameters(cliargs: Namespace) -> int:
    logger.info("Fetching parameters from parameter store.", config=cliargs.config)

    ctx = Context.from_cliargs(cliargs)

    # Fetch parameters from the parameter store
    parameters = ctx.ps_client.get_parameters()

    # write them to the configured env file
    ctx.env_file.update_params(parameters, "main")
    ctx.env_file.write_params()
    return 0



from argparse import ArgumentParser, Namespace
from importlib import metadata
from pathlib import Path
from types import ModuleType
from typing import List

from psenv.core.paths import PSENV_CONFIG_FILE_PATH

from . import config, exec, get, inject, put

COMMANDS = [config, get, put, exec, inject]


def commands() -> List[ModuleType]:
    return sorted(COMMANDS, key=lambda x: x.__name__)


def parse_args() -> Namespace:
    parser = ArgumentParser()
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"psenv :: {metadata.version('pytoolbelt-psenv')} :: AWS Parameter Store Environment Manager",
    )

    parser.add_argument(
        "-c",
        "--config",
        help="Path to the psenv configuration file.",
        type=Path,
        required=False,
        default=PSENV_CONFIG_FILE_PATH,
        metavar="",
    )

    subparsers = parser.add_subparsers(dest="command")
    subparsers.required = True

    for command in commands():
        command.configure_parser(subparsers)

    return parser.parse_args()



class PsenvParameterStoreError(Exception):
    pass


class PsenvConfigNotFoundError(Exception):
    pass


class PsenvConfigError(Exception):
    pass


class PsenvInternalError(Exception):
    pass


class PsenvInvalidAccountError(Exception):
    pass


class PsenvCLIError(Exception):
    pass



from argparse import REMAINDER, Namespace
from typing import Any

import structlog

from psenv.core.command import Command
from psenv.core.context import Context

logger = structlog.get_logger(__name__)


def configure_parser(subparser: Any) -> None:
    put_parser = subparser.add_parser(
        name="exec", description="Run a parameterized command", help="Execute a command with parameters from the parameter store."
    )
    put_parser.set_defaults(func=exec_command)

    put_parser.add_argument("-e", "--env", type=str, required=True, help="The environment to run the command for.")
    put_parser.add_argument("command", nargs=REMAINDER, help="Command to execute with its arguments")


def exec_command(cliargs: Namespace) -> None:
    ctx = Context.from_cliargs(cliargs)
    env = ctx.ps_client.get_parameters()

    with Command(cliargs.command, env=env):
        print(f"Executing command: {cliargs.command} in environment: {env}")
        pass



from psenv.cli import parse_args
from psenv.error_handling import handle_cli_errors


@handle_cli_errors
def main() -> int:
    cliargs = parse_args()
    return cliargs.func(cliargs=cliargs)


if __name__ == "__main__":
    exit(main())



import os
from subprocess import PIPE, Popen
from typing import Dict, List, Optional


class Command(Popen):
    def __init__(self, command: List[str], env: Optional[Dict[str, str]] = None) -> None:
        if command[0] == "--":
            command.pop(0)

        super().__init__(
            command,
            env=env or os.environ.copy(),  # Use provided env or inherit from parent
            stdout=None,  # None means inherit from parent process
            stderr=PIPE,  # Capture stderr for error checking
            text=True,  # Return strings instead of bytes
            bufsize=1,  # Line buffered
            universal_newlines=True,
        )
        self.command = command
        self.env = env

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Get stderr output and return code
        stderr = self.stderr.read() if self.stderr else ""
        return_code = self.wait()

        if return_code != 0:
            self.stderr.close()
            raise RuntimeError(f"Command failed with code {return_code}. Error: {stderr}")

        self.stderr.close()
        return False  # Don't suppress exceptions
